@@ -1,5 +1,5 @@
 //
-// Copyright (C) 2010-2018 Codership Oy <info@codership.com>
+// Copyright (C) 2010-2023 Codership Oy <info@codership.com>
 //
 
 #include "certification.hpp"
@@ -10,7 +10,6 @@
 #include <map>
 #include <algorithm> // std::for_each
 
-
 using namespace galera;
 
 static const bool cert_debug_on(false);
@@ -42,12 +41,13 @@ static std::string const CERT_PARAM_LENGTH_CHECK_DEFAULT("127");
 void
 galera::Certification::register_params(gu::Config& cnf)
 {
-    cnf.add(CERT_PARAM_LOG_CONFLICTS, CERT_PARAM_LOG_CONFLICTS_DEFAULT);
-    cnf.add(CERT_PARAM_OPTIMISTIC_PA, CERT_PARAM_OPTIMISTIC_PA_DEFAULT);
+    const int flags(gu::Config::Flag::type_bool);
+    cnf.add(CERT_PARAM_LOG_CONFLICTS, CERT_PARAM_LOG_CONFLICTS_DEFAULT, flags);
+    cnf.add(CERT_PARAM_OPTIMISTIC_PA, CERT_PARAM_OPTIMISTIC_PA_DEFAULT, flags);
     /* The defaults below are deliberately not reflected in conf: people
      * should not know about these dangerous setting unless they read RTFM. */
-    cnf.add(CERT_PARAM_MAX_LENGTH);
-    cnf.add(CERT_PARAM_LENGTH_CHECK);
+    cnf.add(CERT_PARAM_MAX_LENGTH, gu::Config::Flag::hidden);
+    cnf.add(CERT_PARAM_LENGTH_CHECK, gu::Config::Flag::hidden);
 }
 
 /* a function to get around unset defaults in ctor initialization list */
@@ -70,6 +70,45 @@ length_check(const gu::Config& conf)
         return gu::Config::from_config<int>(CERT_PARAM_LENGTH_CHECK_DEFAULT);
 }
 
+static void
+report_stale_entry(const galera::Certification::CertIndexNG::value_type& ke,
+                   const galera::KeySetIn& key_set)
+{
+    std::cerr << "Found stale entry for key: " << ke->key() << "\n";
+    key_set.rewind();
+    std::cerr << "Key set\n";
+    for (long i = 0; i < key_set.count(); ++i)
+    {
+        const auto& kp = key_set.next();
+        std::cerr << kp << "\n";
+    }
+}
+
+// Verify that there are no stale entries of ts left after index purge.
+// Is stale entry is found, the corresponding key and the key set is
+// printed into stderr. Debug build will assert.
+//
+// This method requires iterating over whole index, so it is relatively
+// expensive, and should be used only for debugging purposes.
+static void
+check_purge_complete(const galera::Certification::CertIndexNG& cert_index,
+                     const galera::TrxHandleSlave* ts,
+                     const galera::KeySetIn& key_set)
+{
+    std::for_each(
+        cert_index.begin(), cert_index.end(),
+        [&cert_index, &key_set,
+         ts](const galera::Certification::CertIndexNG::value_type& ke) {
+            ke->for_each_ref([&ke, &key_set, ts](const TrxHandleSlave* ref) {
+                if (ts == ref)
+                {
+                    report_stale_entry(ke, key_set);
+                }
+                assert(ts != ref);
+            });
+        });
+}
+
 // Purge key set from given index
 static void purge_key_set(galera::Certification::CertIndexNG& cert_index,
                           galera::TrxHandleSlave*             ts,
@@ -102,6 +141,10 @@ static void purge_key_set(galera::Certification::CertIndexNG& cert_index,
             }
         }
     }
+    if (cert_debug_on)
+    {
+        check_purge_complete(cert_index, ts, key_set);
+    }
 }
 
 void
@@ -148,7 +191,7 @@ check_against(const galera::KeyEntryNG*   const found,
 
     bool conflict(false);
 
-    if (gu_likely(0 != ref_trx))
+    if (0 != ref_trx)
     {
         if ((REF_KEY_TYPE == WSREP_KEY_EXCLUSIVE ||
              REF_KEY_TYPE == WSREP_KEY_UPDATE) && ref_trx)
@@ -187,10 +230,7 @@ check_against(const galera::KeyEntryNG*   const found,
             }
             /* fall through */
         case DEPENDENCY:
-            if (conflict)
-                depends_seqno = WSREP_SEQNO_UNDEFINED;
-            else
-                depends_seqno = std::max(ref_trx->global_seqno(), depends_seqno);
+            depends_seqno = std::max(ref_trx->global_seqno(), depends_seqno);
             /* fall through */
         case NOTHING:;
         }
@@ -206,6 +246,7 @@ certify_and_depend_v3to5(const galera::KeyEntryNG*   const found,
                          galera::TrxHandleSlave*     const trx,
                          bool                        const log_conflict)
 {
+    bool ret(false);
     wsrep_seqno_t depends_seqno(trx->depends_seqno());
     wsrep_key_type_t const key_type(key.wsrep_type(trx->version()));
 
@@ -231,54 +272,43 @@ certify_and_depend_v3to5(const galera::KeyEntryNG*   const found,
         check_against<WSREP_KEY_UPDATE>
         (found, key, key_type, trx, log_conflict, depends_seqno) ||
         (key_type >= WSREP_KEY_UPDATE &&
-         /* exclusive keys must be checked against shared */
+         /* exclusive and update keys must be checked against shared */
          (check_against<WSREP_KEY_REFERENCE>
           (found, key, key_type, trx, log_conflict, depends_seqno) ||
           check_against<WSREP_KEY_SHARED>
           (found, key, key_type, trx, log_conflict, depends_seqno))))
     {
-        return true;
-    }
-    else
-    {
-        if (depends_seqno > trx->depends_seqno())
-            trx->set_depends_seqno(depends_seqno);
-        return false;
+        ret = true;
     }
+
+    if (depends_seqno > trx->depends_seqno())
+        trx->set_depends_seqno(depends_seqno);
+
+    return ret;
 }
 
 /* returns true on collision, false otherwise */
 static bool
-certify_v3to5(galera::Certification::CertIndexNG& cert_index_ng,
+certify_v3to5(const galera::Certification::CertIndexNG& cert_index_ng,
               const galera::KeySet::KeyPart&      key,
               galera::TrxHandleSlave*     const   trx,
-              bool                        const   store_keys,
               bool                        const   log_conflicts)
 {
     galera::KeyEntryNG ke(key);
-    galera::Certification::CertIndexNG::iterator ci(cert_index_ng.find(&ke));
+    galera::Certification::CertIndexNG::const_iterator ci(cert_index_ng.find(&ke));
 
     if (cert_index_ng.end() == ci)
     {
-        if (store_keys)
-        {
-            galera::KeyEntryNG* const kep(new galera::KeyEntryNG(ke));
-            ci = cert_index_ng.insert(kep).first;
-
-            cert_debug << "created new entry";
-        }
-        return false;
+        return false; // No match
     }
-    else
-    {
-        cert_debug << "found existing entry";
 
-        galera::KeyEntryNG* const kep(*ci);
-        // Note: For we skip certification for isolated trxs, only
-        // cert index and key_list is populated.
-        return (!trx->is_toi() &&
-                certify_and_depend_v3to5(kep, key, trx, log_conflicts));
-    }
+    cert_debug << "found existing entry";
+
+    galera::KeyEntryNG* const kep(*ci);
+    // Note: For we skip certification for isolated trxs, only
+    // cert index and key_list is populated.
+    return (!trx->is_toi() &&
+            certify_and_depend_v3to5(kep, key, trx, log_conflicts));
 }
 
 // Add key to trx references for trx that passed certification.
@@ -301,62 +331,16 @@ static void do_ref_keys(galera::Certification::CertIndexNG& cert_index,
 
         if (ci == cert_index.end())
         {
-            gu_throw_fatal << "could not find key '" << k
-                           << "' from cert index";
+            galera::KeyEntryNG* const kep(new galera::KeyEntryNG(ke));
+            ci = cert_index.insert(kep).first;
+            cert_debug << "created new entry";
         }
         (*ci)->ref(k.wsrep_type(trx->version()), k, trx);
     }
 }
 
-// Clean up keys from index that were added by trx that failed
-// certification.
-//
-// @param cert_index certification inde
-// @param key_set    key_set used in certification
-// @param processed  number of keys that were processed in certification
-static void do_clean_keys(galera::Certification::CertIndexNG& cert_index,
-                          const galera::TrxHandleSlave* const trx,
-                          const galera::KeySetIn&             key_set,
-                          const long                          processed)
-{
-    /* 'strictly less' comparison is essential in the following loop:
-     * processed key failed cert and was not added to index */
-    for (long i(0); i < processed; ++i)
-    {
-        KeyEntryNG ke(key_set.next());
-
-        // Clean up cert index from entries which were added by this trx
-        galera::Certification::CertIndexNG::iterator ci(cert_index.find(&ke));
-
-        if (gu_likely(ci != cert_index.end()))
-        {
-            galera::KeyEntryNG* kep(*ci);
-
-            if (kep->referenced() == false)
-            {
-                // kel was added to cert_index_ by this trx -
-                // remove from cert_index_ and fall through to delete
-                cert_index.erase(ci);
-            }
-            else continue;
-
-            assert(kep->referenced() == false);
-
-            delete kep;
-        }
-        else if(ke.key().wsrep_type(trx->version()) == WSREP_KEY_SHARED)
-        {
-            assert(0); // we actually should never be here, the key should
-                       // be either added to cert_index_ or be there already
-            log_warn  << "could not find shared key '"
-                      << ke.key() << "' from cert index";
-        }
-        else { /* non-shared keys can duplicate shared in the key set */ }
-    }
-}
-
 galera::Certification::TestResult
-galera::Certification::do_test_v3to5(TrxHandleSlave* trx, bool store_keys)
+galera::Certification::do_test_v3to5(TrxHandleSlave* trx)
 {
     cert_debug << "BEGIN CERTIFICATION v" << trx->version() << ": " << *trx;
 
@@ -376,24 +360,23 @@ galera::Certification::do_test_v3to5(TrxHandleSlave* trx, bool store_keys)
     {
         const KeySet::KeyPart& key(key_set.next());
 
-        if (certify_v3to5(cert_index_ng_, key, trx, store_keys, log_conflicts_))
+        if (certify_v3to5(cert_index_ng_, key, trx, log_conflicts_))
         {
+            trx->set_depends_seqno(std::max(trx->depends_seqno(), last_pa_unsafe_));
             goto cert_fail;
         }
     }
 
     trx->set_depends_seqno(std::max(trx->depends_seqno(), last_pa_unsafe_));
 
-    if (store_keys == true)
-    {
-        assert (key_count == processed);
-        key_set.rewind();
-        do_ref_keys(cert_index_ng_, trx, key_set, key_count);
+    assert (key_count == processed);
+    key_set.rewind();
+    do_ref_keys(cert_index_ng_, trx, key_set, key_count);
 
-        if (trx->pa_unsafe()) last_pa_unsafe_ = trx->global_seqno();
+    if (trx->pa_unsafe()) last_pa_unsafe_ = trx->global_seqno();
+
+    key_count_ += key_count;
 
-        key_count_ += key_count;
-    }
     cert_debug << "END CERTIFICATION (success): " << *trx;
     return TEST_OK;
 
@@ -402,14 +385,7 @@ cert_fail:
     cert_debug << "END CERTIFICATION (failed): " << *trx;
 
     assert (processed < key_count);
-
-    if (store_keys == true)
-    {
-        /* Clean up key entries allocated for this trx */
-        key_set.rewind();
-        do_clean_keys(cert_index_ng_, trx, key_set, processed);
-        assert(cert_index_ng_.size() == prev_cert_index_size);
-    }
+    assert(cert_index_ng_.size() == prev_cert_index_size);
 
     return TEST_FAILED;
 }
@@ -433,8 +409,9 @@ trx_cert_version_match(int const trx_version, int const cert_version)
 }
 
 galera::Certification::TestResult
-galera::Certification::do_test(const TrxHandleSlavePtr& trx, bool store_keys)
+galera::Certification::do_test(const TrxHandleSlavePtr& trx)
 {
+    assert(mutex_.owned());
     assert(trx->source_id() != WSREP_UUID_UNDEFINED);
 
     if (!trx_cert_version_match(trx->version(), version_))
@@ -463,8 +440,6 @@ galera::Certification::do_test(const TrxHandleSlavePtr& trx, bool store_keys)
 
     TestResult res(TEST_FAILED);
 
-    gu::Lock lock(mutex_); // why do we need that? - e.g. set_trx_committed()
-
     /* initialize parent seqno */
     if (gu_unlikely(trx_map_.empty()))
     {
@@ -488,16 +463,14 @@ galera::Certification::do_test(const TrxHandleSlavePtr& trx, bool store_keys)
     case 3:
     case 4:
     case 5:
-        res = do_test_v3to5(trx.get(), store_keys);
+        res = do_test_v3to5(trx.get());
         break;
     default:
         gu_throw_fatal << "certification test for version "
                        << version_ << " not implemented";
     }
 
-    assert(TEST_FAILED == res || trx->depends_seqno() >= 0);
-
-    if (store_keys == true && res == TEST_OK)
+    if (res == TEST_OK)
     {
         ++trx_count_;
         gu::Lock lock(stats_mutex_);
@@ -567,13 +540,11 @@ galera::Certification::do_test_preordered(TrxHandleSlave* trx)
 
 // Prepare a copy of TrxHandleSlave with private storage
 galera::NBOEntry copy_ts(
+    gcache::GCache& gcache,
     galera::TrxHandleSlave* ts,
     galera::TrxHandleSlave::Pool& pool,
     gu::shared_ptr<NBOCtx>::type nbo_ctx)
 {
-    // FIXME: Pass proper working directory from config to MappedBuffer ctor
-    gu::shared_ptr<galera::MappedBuffer>::type buf(
-        new galera::MappedBuffer("/tmp"));
     assert(ts->action().first && ts->action().second);
     if (ts->action().first == 0)
     {
@@ -582,30 +553,33 @@ galera::NBOEntry copy_ts(
             << *ts;
     }
 
+    gu::shared_ptr<NBOEntry::Buffer>::type buf(new NBOEntry::Buffer);
     buf->resize(ts->action().second);
-    std::copy(static_cast<const gu::byte_t*>(ts->action().first),
-              static_cast<const gu::byte_t*>(ts->action().first)
-              + ts->action().second,
-              buf->begin());
+    if (buf->size() > size_t(std::numeric_limits<int32_t>::max()))
+        gu_throw_error(ERANGE) << "Buffer size " << buf->size()
+                               << " out of range";
+
+    const gu::byte_t* const plaintext
+        (static_cast<const gu::byte_t*>
+         (gcache.get_ro_plaintext(ts->action().first)));
+    std::copy(plaintext, plaintext + ts->action().second, buf->begin());
+    gcache.drop_plaintext(ts->action().first);
 
     galera::TrxHandleSlaveDeleter d;
     gu::shared_ptr<galera::TrxHandleSlave>::type new_ts(
         galera::TrxHandleSlave::New(ts->local(), pool), d);
-    if (buf->size() > size_t(std::numeric_limits<int32_t>::max()))
-        gu_throw_error(ERANGE) << "Buffer size " << buf->size()
-                               << " out of range";
     gcs_action act = {ts->global_seqno(), ts->local_seqno(),
                       &(*buf)[0], static_cast<int32_t>(buf->size()),
                       GCS_ACT_WRITESET};
     if (ts->certified() == false)
     {
         // TrxHandleSlave is from group
-        gu_trace(new_ts->unserialize<true>(act));
+        gu_trace((new_ts->unserialize<true,false>(gcache,act)));
     }
     else
     {
         // TrxHandleSlave is from IST
-        gu_trace(new_ts->unserialize<false>(act));
+        gu_trace((new_ts->unserialize<false,false>(gcache,act)));
     }
     new_ts->set_local(ts->local());
     return galera::NBOEntry(new_ts, buf, nbo_ctx);
@@ -842,8 +816,8 @@ galera::Certification::TestResult galera::Certification::do_test_nbo(
         log_debug << "NBO start: " << *ts;
         // We need a copy of ts since the lifetime of NBO may exceed
         // the lifetime of the buffer in GCache
-        NBOEntry entry(copy_ts(ts.get(), nbo_pool_, nbo_ctx_unlocked(
-                                   ts->global_seqno())));
+        NBOEntry entry(copy_ts(gcache_, ts.get(), nbo_pool_,
+                               nbo_ctx_unlocked(ts->global_seqno())));
 
         TrxHandleSlave* new_ts(entry.ts_ptr());
         const KeySetIn& key_set(new_ts->write_set().keyset());
@@ -932,10 +906,13 @@ galera::Certification::TestResult galera::Certification::do_test_nbo(
     return ret;
 }
 
-galera::Certification::Certification(gu::Config& conf, ServiceThd* thd)
+galera::Certification::Certification(gu::Config&     conf,
+                                     gcache::GCache& cache,
+                                     ServiceThd*     thd)
     :
     version_               (-1),
     conf_                  (conf),
+    gcache_                (cache),
     trx_map_               (),
     cert_index_ng_         (),
     nbo_map_               (),
@@ -945,7 +922,7 @@ galera::Certification::Certification(gu::Config& conf, ServiceThd* thd)
     deps_set_              (),
     current_view_          (),
     service_thd_           (thd),
-    mutex_                 (),
+    mutex_                 (gu::get_mutex_key(gu::GU_MUTEX_KEY_CERTIFICATION)),
     trx_size_warn_count_   (0),
     initial_position_      (-1),
     position_              (-1),
@@ -954,7 +931,7 @@ galera::Certification::Certification(gu::Config& conf, ServiceThd* thd)
     last_pa_unsafe_        (-1),
     last_preordered_seqno_ (position_),
     last_preordered_id_    (0),
-    stats_mutex_           (),
+    stats_mutex_           (gu::get_mutex_key(gu::GU_MUTEX_KEY_CERTIFICATION_STATS)),
     n_certified_           (0),
     deps_dist_             (0),
     cert_interval_         (0),
@@ -1124,33 +1101,26 @@ galera::Certification::adjust_position(const View&         view,
     }
 }
 
-wsrep_seqno_t
-galera::Certification::increment_position()
-{
-    gu::Lock lock(mutex_);
-    position_++;
-    return position_;
-}
-
 galera::Certification::TestResult
-galera::Certification::test(const TrxHandleSlavePtr& trx, bool store_keys)
+galera::Certification::test(const TrxHandleSlavePtr& trx)
 {
+    assert(mutex_.owned());
     assert(trx->global_seqno() >= 0 /* && trx->local_seqno() >= 0 */);
 
-    const TestResult ret
-        (trx->preordered() ?
-         do_test_preordered(trx.get()) : do_test(trx, store_keys));
+    const TestResult ret(trx->preordered() ? do_test_preordered(trx.get())
+                                           : do_test(trx));
 
-    assert(TEST_FAILED == ret || trx->depends_seqno() >= 0);
-
-    if (gu_unlikely(ret != TEST_OK)) { trx->mark_dummy(); }
+    if (gu_unlikely(ret != TEST_OK))
+    {
+        trx->mark_dummy();
+    }
 
     return ret;
 }
 
-
 wsrep_seqno_t galera::Certification::get_safe_to_discard_seqno_() const
 {
+    assert(mutex_.owned());
     wsrep_seqno_t retval;
     if (deps_set_.empty() == true)
     {
@@ -1169,10 +1139,14 @@ galera::Certification::purge_trxs_upto_(wsrep_seqno_t const seqno,
                                         bool const          handle_gcache)
 {
     assert (seqno > 0);
+    assert(mutex_.owned());
 
     TrxMap::iterator purge_bound(trx_map_.upper_bound(seqno));
 
-    cert_debug << "purging index up to " << seqno;
+    cert_debug << "purging index up to " << seqno << ", safe to discard seqno " << get_safe_to_discard_seqno_();
+
+    assert(purge_bound == trx_map_.end() ||
+           purge_bound->first <= get_safe_to_discard_seqno_() + 1);
 
     for_each(trx_map_.begin(), purge_bound, PurgeAndDiscard(*this));
     trx_map_.erase(trx_map_.begin(), purge_bound);
@@ -1194,13 +1168,13 @@ galera::Certification::TestResult
 galera::Certification::append_trx(const TrxHandleSlavePtr& trx)
 {
 // explicit ROLLBACK is dummy()    assert(!trx->is_dummy());
-    assert(trx->global_seqno() >= 0 /* && trx->local_seqno() >= 0 */);
+    assert(trx->global_seqno() > 0 /* && trx->local_seqno() >= 0 */);
     assert(trx->global_seqno() > position_);
 
 #ifndef NDEBUG
     bool const explicit_rollback(trx->explicit_rollback());
 #endif /* NDEBUG */
-
+    TestResult retval = TEST_FAILED;
     {
         gu::Lock lock(mutex_);
 
@@ -1239,19 +1213,14 @@ galera::Certification::append_trx(const TrxHandleSlavePtr& trx)
             }
             else
             {
-                cert_debug << "purging index up to " << trim_seqno;
+                cert_debug << "append_trx: purging index up to " << trim_seqno;
             }
 
             purge_trxs_upto_(trim_seqno, true);
         }
-    }
 
-    const TestResult retval(test(trx, true));
+        retval = test(trx);
 
-    {
-        assert(trx->global_seqno() > 0);
-
-        gu::Lock lock(mutex_);
         if (trx_map_.insert(
                 std::make_pair(trx->global_seqno(), trx)).second == false)
             gu_throw_fatal << "duplicate trx entry " << *trx;
@@ -1281,6 +1250,22 @@ galera::Certification::append_trx(const TrxHandleSlavePtr& trx)
 }
 
 
+void galera::Certification::append_dummy_preload(const TrxHandleSlavePtr& trx)
+{
+    assert(trx->global_seqno() >= 0);
+    assert(trx->global_seqno() > position_);
+    gu::Lock lock(mutex_);
+    /* Dummy preloads have only meta data available, not the whole write set,
+       so modifying or accessing the trx object causes problems later on.
+       Insert nullptr as a placeholder for seqno. */
+    if (not trx_map_.insert(std::make_pair(trx->global_seqno(), nullptr))
+                .second)
+    {
+        gu_throw_fatal << "duplicate trx entry in dummy preload";
+    }
+    position_ = trx->global_seqno();
+}
+
 wsrep_seqno_t galera::Certification::set_trx_committed(TrxHandleSlave& trx)
 {
     assert(trx.global_seqno() >= 0);
