@@ -1,5 +1,5 @@
 //
-// Copyright (C) 2010-2020 Codership Oy <info@codership.com>
+// Copyright (C) 2010-2021 Codership Oy <info@codership.com>
 //
 
 //! @file replicator_smm.hpp
@@ -11,6 +11,7 @@
 #define GALERA_REPLICATOR_SMM_HPP
 
 #include "replicator.hpp"
+#include "progress_callback.hpp"
 
 #include "gu_init.h"
 #include "GCache.hpp"
@@ -57,6 +58,11 @@ namespace galera
         int trx_proto_ver() const { return trx_params_.version_; }
         int repl_proto_ver() const{ return protocol_version_; }
 
+        wsrep_status_t enc_set_key(const wsrep_enc_key_t& key)
+        {
+            gcache_.set_enc_key(key); return WSREP_OK;
+        }
+
         wsrep_status_t connect(const std::string& cluster_name,
                                const std::string& cluster_url,
                                const std::string& state_donor,
@@ -142,7 +148,9 @@ namespace galera
         void process_trx(void* recv_ctx, const TrxHandleSlavePtr& trx);
         void process_commit_cut(wsrep_seqno_t seq, wsrep_seqno_t seqno_l);
         void submit_view_info(void* recv_ctx, const wsrep_view_info_t* cc);
-        void process_conf_change(void* recv_ctx, const struct gcs_action& cc);
+        void process_conf_change(void*                    recv_ctx,
+                                 const gcs_act_cchange&   cc,
+                                 const struct gcs_action& act);
         void process_state_req(void* recv_ctx, const void* req,
                                size_t req_size, wsrep_seqno_t seqno_l,
                                wsrep_seqno_t donor_seq);
@@ -154,6 +162,18 @@ namespace galera
         void                          stats_reset();
         void                          stats_free(struct wsrep_stats_var*);
 
+        wsrep_status_t fetch_pfs_info(wsrep_node_info_t** nodes,
+                                      uint32_t* size,
+                                      int32_t* my_index,
+                                      uint32_t max_version);
+        wsrep_status_t fetch_pfs_stat(wsrep_node_stat_t** nodes,
+                                      uint32_t* size,
+                                      int32_t* my_index,
+                                      uint32_t max_version);
+
+        void free_pfs_info(wsrep_node_info_t* nodes);
+        void free_pfs_stat(wsrep_node_stat_t* nodes);
+
         /*! @throws NotFound */
         void           set_param (const std::string& key,
                                   const std::string& value);
@@ -180,7 +200,8 @@ namespace galera
         // IST Action handler interface
         void ist_trx(const TrxHandleSlavePtr& ts, bool must_apply,
                      bool preload);
-        void ist_cc(const gcs_action&, bool must_apply, bool preload);
+        void ist_cc(const gcs_act_cchange&, const gcs_action&, bool must_apply,
+                    bool preload);
         void ist_end(int error);
 
         // Cancel local and enter apply monitors for TrxHandle
@@ -256,8 +277,8 @@ namespace galera
         public:
             ISTEventQueue()
                 :
-                mutex_(),
-                cond_(),
+                mutex_(gu::get_mutex_key(gu::GU_MUTEX_KEY_IST_EVENT_QUEUE)),
+                cond_(gu::get_cond_key(gu::GU_COND_KEY_IST_EVENT_QUEUE)),
                 eof_(false),
                 error_(0),
                 queue_()
@@ -347,6 +368,9 @@ namespace galera
 
         bool corrupt() const { return st_.corrupt(); }
 
+        wsrep_status_t get_membership(wsrep_allocator_cb        alloc,
+                                      struct wsrep_membership** memb) const;
+
         struct InitConfig
         {
             InitConfig(gu::Config&, const char* node_addr,const char* base_dir);
@@ -396,19 +420,21 @@ namespace galera
         static wsrep_cap_t capabilities(int protocol_version);
 
         // Return the global seqno of the last transaction which has
-        // released commit order. Note that this does not mean that
-        // the transaction with given gtid has completed the commit
-        // on application side.
+        // commmitted.
+        //
+        // galera-bugs#555: Assign last seen from apply_monitor_.last_left()
+        // because apply monitor is held until the whole transaction is over.
+        // Commit monitor may be released early due to group commit.
         wsrep_seqno_t last_committed()
         {
-            return co_mode_ != CommitOrder::BYPASS ?
-                   commit_monitor_.last_left() : apply_monitor_.last_left();
+            return apply_monitor_.last_left();
         }
 
         void report_last_committed(wsrep_seqno_t purge_seqno)
         {
             if (gu_unlikely(purge_seqno != -1))
             {
+                assert(purge_seqno <= last_committed());
                 service_thd_.report_last_committed(purge_seqno);
             }
         }
@@ -463,8 +489,6 @@ namespace galera
                                          const TrxHandleSlavePtr&);
         wsrep_status_t cert_and_catch   (TrxHandleMaster*,
                                          const TrxHandleSlavePtr&);
-        wsrep_status_t cert_for_aborted (const TrxHandleSlavePtr&);
-
         // Enter apply monitor for local transaction. Return true
         // if apply monitor was grabbed.
         bool enter_apply_monitor_for_local(TrxHandleMaster&,
@@ -493,27 +517,45 @@ namespace galera
 
             explicit
             LocalOrder(const TrxHandleSlave& ts)
-                :
-                seqno_(ts.local_seqno())
-#if defined(GU_DBUG_ON) || !defined(NDEBUG)
-                ,trx_(&ts)
-#endif //GU_DBUG_ON
+                : seqno_(ts.local_seqno())
+                , cond_(&ts.local_order_cond_)
+                , trx_(&ts)
             { }
 
             LocalOrder(wsrep_seqno_t seqno, const TrxHandleSlave* ts = NULL)
-                :
-                seqno_(seqno)
-#if defined(GU_DBUG_ON) || !defined(NDEBUG)
-                ,trx_(ts)
-#endif //GU_DBUG_ON
+                : seqno_(seqno)
+                , cond_(ts ? &ts->local_order_cond_ :
+                        new gu::Cond(
+                            gu::get_cond_key(gu::GU_COND_KEY_LOCAL_MONITOR)))
+                , trx_(ts)
             {
-#if defined(GU_DBUG_ON) || !defined(NDEBUG)
+                assert(ts || cond_);
                 assert((trx_ && seqno_ == trx_->local_seqno()) || !trx_);
-#endif //GU_DBUG_ON
             }
 
+            ~LocalOrder()
+            {
+                if (not trx_) delete cond_;
+            }
+
+#if defined(GU_DBUG_ON) || !defined(NDEBUG)
+            LocalOrder(const LocalOrder& other)
+                : seqno_(other.seqno_)
+                  // Do not copy cond for debug diagnostics as it may be
+                  // deleted in destructor, which would cause double free.
+                , cond_()
+                , trx_(other.trx_)
+            { }
+
+#else
+            LocalOrder(const LocalOrder&) = delete;
+#endif
+            LocalOrder& operator=(const LocalOrder&) = delete;
+
             wsrep_seqno_t seqno() const { return seqno_; }
 
+            gu::Cond* cond() { return cond_; }
+
             bool condition(wsrep_seqno_t last_entered,
                            wsrep_seqno_t last_left) const
             {
@@ -538,17 +580,21 @@ namespace galera
                         mutex.lock();
                     }
                 }
+                else
+                {
+                    mutex.unlock();
+                    GU_DBUG_SYNC_WAIT("local_monitor_enter_sync");
+                    mutex.lock();
+                }
             }
 #endif //GU_DBUG_ON
 
 #ifndef NDEBUG
             LocalOrder()
-                :
-                seqno_(WSREP_SEQNO_UNDEFINED)
-#if defined(GU_DBUG_ON) || !defined(NDEBUG)
-                ,trx_(NULL)
-#endif /* GU_DBUG_ON || !NDEBUG */
-            {}
+                : seqno_(WSREP_SEQNO_UNDEFINED)
+                , cond_()
+                , trx_()
+            { }
 #endif /* NDEBUG */
 
             void print(std::ostream& os) const
@@ -557,15 +603,11 @@ namespace galera
             }
 
         private:
-#ifdef NDEBUG
-            LocalOrder(const LocalOrder& o);
-#endif /* NDEBUG */
             wsrep_seqno_t const seqno_;
-#if defined(GU_DBUG_ON) || !defined(NDEBUG)
-            // this pointer is for debugging purposes only and
-            // is not guaranteed to point at a valid location
+            gu::Cond* cond_;
+            // This pointer is not guaranteed to point into valid location
+            // if the object was copy-constructed.
             const TrxHandleSlave* const trx_;
-#endif /* GU_DBUG_ON || !NDEBUG */
         };
 
         class ApplyOrder
@@ -576,6 +618,7 @@ namespace galera
                 :
                 global_seqno_ (ts.global_seqno()),
                 depends_seqno_(ts.depends_seqno()),
+                cond_(&ts.apply_order_cond_),
                 is_local_     (ts.local()),
                 is_toi_       (ts.is_toi())
 #ifndef NDEBUG
@@ -593,6 +636,7 @@ namespace galera
                 :
                 global_seqno_ (gs),
                 depends_seqno_(ds),
+                cond_(),
                 is_local_     (l),
                 is_toi_       (false)
 #ifndef NDEBUG
@@ -602,6 +646,8 @@ namespace galera
 
             wsrep_seqno_t seqno() const { return global_seqno_; }
 
+            gu::Cond* cond() { return cond_; }
+
             bool condition(wsrep_seqno_t last_entered,
                            wsrep_seqno_t last_left) const
             {
@@ -632,6 +678,7 @@ namespace galera
                 :
                 global_seqno_ (WSREP_SEQNO_UNDEFINED),
                 depends_seqno_(WSREP_SEQNO_UNDEFINED),
+                cond_(),
                 is_local_     (false),
                 is_toi_       (false),
                 trx_          (NULL)
@@ -651,6 +698,7 @@ namespace galera
 #endif /* NDEBUG */
             const wsrep_seqno_t global_seqno_;
             const wsrep_seqno_t depends_seqno_;
+            gu::Cond* cond_;
             const bool is_local_;
             const bool is_toi_;
 #ifndef NDEBUG
@@ -691,6 +739,7 @@ namespace galera
             CommitOrder(const TrxHandleSlave& ts, Mode mode)
                 :
                 global_seqno_(ts.global_seqno()),
+                cond_(&ts.commit_order_cond_),
                 mode_(mode),
                 is_local_(ts.local())
 #ifndef NDEBUG
@@ -701,6 +750,7 @@ namespace galera
             CommitOrder(wsrep_seqno_t gs, Mode mode, bool local = false)
                 :
                 global_seqno_(gs),
+                cond_(),
                 mode_(mode),
                 is_local_(local)
 #ifndef NDEBUG
@@ -710,6 +760,8 @@ namespace galera
 
             wsrep_seqno_t seqno() const { return global_seqno_; }
 
+            gu::Cond* cond() { return cond_; }
+
             bool condition(wsrep_seqno_t last_entered,
                            wsrep_seqno_t last_left) const
             {
@@ -751,6 +803,7 @@ namespace galera
             CommitOrder()
                 :
                 global_seqno_ (WSREP_SEQNO_UNDEFINED),
+                cond_         (),
                 mode_         (OOOC),
                 is_local_     (false),
                 trx_          (NULL)
@@ -770,6 +823,7 @@ namespace galera
             CommitOrder(const CommitOrder&);
 #endif
             const wsrep_seqno_t global_seqno_;
+            gu::Cond* cond_;
             const Mode mode_;
             const bool is_local_;
 #ifndef NDEBUG
@@ -982,13 +1036,16 @@ namespace galera
         bool          sst_received_;
 
         // services
-        gcache::GCache gcache_;
-        GCS_IMPL       gcs_;
-        ServiceThd     service_thd_;
+        ProgressCallback<int64_t> gcache_progress_cb_;
+        gcache::GCache   gcache_;
+        ProgressCallback<gcs_seqno_t> joined_progress_cb_;
+        GCS_IMPL         gcs_;
+        ServiceThd       service_thd_;
 
         // action sources
         TrxHandleSlave::Pool slave_pool_;
         ActionSource*        as_;
+        ProgressCallback<wsrep_seqno_t>ist_progress_cb_;
         ist::Receiver        ist_receiver_;
         ist::AsyncSenderMap  ist_senders_;
 
@@ -1000,7 +1057,8 @@ namespace galera
         {
         public:
             PendingCertQueue(gcache::GCache& gcache) :
-                mutex_(),
+                mutex_(gu::get_mutex_key(
+                           gu::GU_MUTEX_KEY_PENDING_CERTIFICATION)),
                 ts_queue_(),
                 gcache_(gcache)
             { }
