@@ -37,11 +37,13 @@ namespace galera
 
         enum Version
         {
-            VER3 = 3
+            VER3 = 3,
+            VER4,
+            VER5
         };
 
         /* Max header version that we can understand */
-        static Version const MAX_VERSION = VER3;
+        static Version const MAX_VERSION = VER5;
 
         /* Parses beginning of the header to detect writeset version and
          * returns it as raw integer for backward compatibility
@@ -87,9 +89,12 @@ namespace galera
             switch (v)
             {
             case VER3: return VER3;
+            case VER4: return VER4;
+            case VER5: return VER5;
             }
 
             gu_throw_error (EPROTO) << "Unrecognized writeset version: " << v;
+	    return VER3; // silence compiler
         }
 
         /* These flags should be fixed to wire protocol version and so
@@ -102,12 +107,28 @@ namespace galera
             F_TOI         = 1 << 2,
             F_PA_UNSAFE   = 1 << 3,
             F_COMMUTATIVE = 1 << 4,
-            F_NATIVE      = 1 << 5
+            F_NATIVE      = 1 << 5,
+            F_BEGIN       = 1 << 6,
+            F_PREPARE     = 1 << 7,
+            /*
+             * reserved for provider extension
+             */
+            F_CERTIFIED   = 1 << 14, // needed to correctly interprete pa_range
+                                     // field (VER5 and up)
+            F_PREORDERED  = 1 << 15  // (VER5 and up)
         };
 
-        /* this takes care of converting wsrep API flags to on-the-wire flags */
-        static uint32_t
-        wsrep_flags_to_ws_flags (uint32_t const flags);
+        static bool const FLAGS_MATCH_API_FLAGS =
+                           (WSREP_FLAG_TRX_END     == F_COMMIT       &&
+                            WSREP_FLAG_ROLLBACK    == F_ROLLBACK     &&
+                            WSREP_FLAG_ISOLATION   == F_TOI          &&
+                            WSREP_FLAG_PA_UNSAFE   == F_PA_UNSAFE    &&
+                            WSREP_FLAG_COMMUTATIVE == F_COMMUTATIVE  &&
+                            WSREP_FLAG_NATIVE      == F_NATIVE       &&
+                            WSREP_FLAG_TRX_START   == F_BEGIN        &&
+                            WSREP_FLAG_TRX_PREPARE == F_PREPARE);
+
+        static uint32_t wsrep_flags_to_ws_flags (uint32_t flags);
 
         typedef gu::RecordSet::GatherVector GatherVector;
 
@@ -131,6 +152,10 @@ namespace galera
                 switch (ver)
                 {
                 case VER3:
+                // fall through
+                case VER4:
+                // fall through
+                case VER5:
                 {
                     GU_COMPILE_ASSERT(0 == (V3_SIZE % GU_MIN_ALIGNMENT),
                                       unaligned_header_size);
@@ -163,17 +188,13 @@ namespace galera
                            GatherVector&          out);
 
             /* records last_seen, timestamp and CRC before replication */
-            void set_last_seen (const wsrep_seqno_t& ls);
+            void finalize(wsrep_seqno_t ls, int pa_range);
 
             /* records partial seqno, pa_range, timestamp and CRC before
              * replication (for preordered events)*/
-            void set_preordered (uint16_t pa_range)
+            void finalize_preordered(uint16_t pa_range)
             {
-                uint16_t* const pa(reinterpret_cast<uint16_t*>
-                                   (ptr_ + V3_PA_RANGE_OFF));
-
-                *pa = gu::htog<uint16_t>(pa_range);
-                set_last_seen (0);
+                finalize(0, pa_range);
             }
 
             /* This is for WriteSetIn */
@@ -197,7 +218,7 @@ namespace galera
             {
                 ver_ = version(buf);
                 ptr_ = static_cast<gu::byte_t*>(const_cast<void*>(buf.ptr));
-                size_ = check_size (ver_, ptr_, buf.size);
+                gu_trace(size_ = check_size (ver_, ptr_, buf.size));
                 Checksum::verify(ver_, ptr_, size_);
             }
 
@@ -242,55 +263,58 @@ namespace galera
 
             uint16_t         flags() const
             {
-                return gu::gtoh(
-                    *(reinterpret_cast<const uint16_t*>(ptr_ + V3_FLAGS_OFF))
-                    );
+                uint16_t ret;
+                gu::unserialize2(ptr_, V3_FLAGS_OFF, ret);
+                return ret;
             }
 
             uint16_t         pa_range() const
             {
-                return gu::gtoh(
-                    *(reinterpret_cast<const uint16_t*>(ptr_ + V3_PA_RANGE_OFF))
-                    );
+                uint16_t ret;
+                gu::unserialize2(ptr_, V3_PA_RANGE_OFF, ret);
+                return ret;
             }
 
             wsrep_seqno_t    last_seen() const
             {
-                assert (pa_range() == 0);
+                assert (pa_range() == 0 || version() >= VER5);
                 return seqno_priv();
             }
 
             wsrep_seqno_t    seqno() const
             {
-                assert (pa_range() > 0);
                 return seqno_priv();
             }
 
             long long        timestamp() const
             {
-                return gu::gtoh(
-                    *(reinterpret_cast<const uint64_t*>(ptr_+ V3_TIMESTAMP_OFF))
-                    );
+                long long ret;
+                gu::unserialize8(ptr_, V3_TIMESTAMP_OFF, ret);
+                return ret;
             }
 
             const wsrep_uuid_t& source_id() const
             {
+                /* This one is tricky. I would not like to create a copy
+                 * of 16 bytes for the sole purpose of referencing it when
+                 * alignment in the buffer is already guaranteed */
+                assert(uintptr_t(ptr_ + V3_SOURCE_ID_OFF)%GU_WORD_BYTES == 0);
                 return *(reinterpret_cast<const wsrep_uuid_t*>
                          (ptr_ + V3_SOURCE_ID_OFF));
             }
 
-            wsrep_trx_id_t conn_id() const
+            wsrep_conn_id_t conn_id() const
             {
-                return gu::gtoh(
-                    *(reinterpret_cast<const uint64_t*>(ptr_ + V3_CONN_ID_OFF))
-                    );
+                wsrep_conn_id_t ret;
+                gu::unserialize8(ptr_, V3_CONN_ID_OFF, ret);
+                return ret;
             }
 
             wsrep_trx_id_t trx_id() const
             {
-                return gu::gtoh(
-                    *(reinterpret_cast<const uint64_t*>(ptr_ + V3_TRX_ID_OFF))
-                    );
+                wsrep_trx_id_t ret;
+                gu::unserialize8(ptr_, V3_TRX_ID_OFF, ret);
+                return ret;
             }
 
             const gu::byte_t* payload() const
@@ -299,7 +323,7 @@ namespace galera
             }
 
             /* to set seqno and parallel applying range after certification */
-            void set_seqno(const wsrep_seqno_t& seqno, uint16_t pa_range);
+            void set_seqno(wsrep_seqno_t seqno, uint16_t pa_range);
 
             gu::Buf copy(bool include_keys, bool include_unrd) const;
 
@@ -415,9 +439,9 @@ namespace galera
 
             wsrep_seqno_t seqno_priv() const
             {
-                return gu::gtoh(
-                    *(reinterpret_cast<const uint64_t*>(ptr_+ V3_LAST_SEEN_OFF))
-                    );
+                wsrep_seqno_t ret;
+                gu::unserialize8(ptr_, V3_LAST_SEEN_OFF, ret);
+                return ret;
             }
 
             static void
@@ -431,47 +455,38 @@ namespace galera
 
     private:
 
-        static bool const WRITESET_FLAGS_MATCH_API_FLAGS =
-                           (WSREP_FLAG_COMMIT      == F_COMMIT       &&
-                            WSREP_FLAG_ROLLBACK    == F_ROLLBACK     &&
-                            WSREP_FLAG_ISOLATION   == F_TOI          &&
-                            WSREP_FLAG_PA_UNSAFE   == F_PA_UNSAFE    &&
-                            WSREP_FLAG_COMMUTATIVE == F_COMMUTATIVE  &&
-                            WSREP_FLAG_NATIVE      == F_NATIVE);
-
         /* this assert should be removed when wsrep API flags become
          * explicitly incompatible with wirteset flags */
-        GU_COMPILE_ASSERT(WRITESET_FLAGS_MATCH_API_FLAGS, flags_incompatible);
+        GU_COMPILE_ASSERT(FLAGS_MATCH_API_FLAGS, flags_incompatible);
 
-        template<bool>
+        template <bool>
         static inline uint32_t
         wsrep_flags_to_ws_flags_tmpl (uint32_t const flags)
         {
+            assert(0); // remove when needed
             uint32_t ret(0);
 
-            if (flags & WSREP_FLAG_COMMIT)      ret |= F_COMMIT;
+            if (flags & WSREP_FLAG_TRX_END)     ret |= F_COMMIT;
             if (flags & WSREP_FLAG_ROLLBACK)    ret |= F_ROLLBACK;
             if (flags & WSREP_FLAG_ISOLATION)   ret |= F_TOI;
             if (flags & WSREP_FLAG_PA_UNSAFE)   ret |= F_PA_UNSAFE;
             if (flags & WSREP_FLAG_COMMUTATIVE) ret |= F_COMMUTATIVE;
             if (flags & WSREP_FLAG_NATIVE)      ret |= F_NATIVE;
+            if (flags & WSREP_FLAG_TRX_START)   ret |= F_BEGIN;
+            if (flags & WSREP_FLAG_TRX_PREPARE) ret |= F_PREPARE;
 
             return ret;
         }
 
     }; /* class WriteSetNG */
 
-    /* specialization for the case when WS flags fully match API flags */
     template <> inline uint32_t
     WriteSetNG::wsrep_flags_to_ws_flags_tmpl<true>(uint32_t const flags)
     { return flags; }
 
     inline uint32_t
     WriteSetNG::wsrep_flags_to_ws_flags (uint32_t const flags)
-    {
-        return wsrep_flags_to_ws_flags_tmpl<WRITESET_FLAGS_MATCH_API_FLAGS>
-            (flags);
-    }
+    { return wsrep_flags_to_ws_flags_tmpl<FLAGS_MATCH_API_FLAGS>(flags); }
 
     class WriteSetOut
     {
@@ -497,7 +512,7 @@ namespace galera
             kbn_   (base_name_),
             keys_  (reserved,
                     (reserved_size >>= 6, reserved_size <<= 3, reserved_size),
-                    kbn_, kver, rsv),
+                    kbn_, kver, rsv, ver),
             /* 5/8 of reserved goes to data set  */
             dbn_   (base_name_),
             data_  (reserved + reserved_size, reserved_size*5, dbn_, dver, rsv),
@@ -557,13 +572,13 @@ namespace galera
 
 
         /* !!! This returns header without checksum! *
-         *     Use set_last_seen() to finalize it.   */
+         *     Use finalize() to finalize it.   */
         size_t gather(const wsrep_uuid_t&       source,
                       const wsrep_conn_id_t&    conn,
                       const wsrep_trx_id_t&     trx,
                       WriteSetNG::GatherVector& out)
         {
-            check_size();
+            gu_trace(check_size());
 
             out->reserve (out->size() + keys_.page_count() + data_.page_count()
                           + unrd_.page_count() + 1 /* global header */);
@@ -585,12 +600,36 @@ namespace galera
             return out_size;
         }
 
-        void set_last_seen (const wsrep_seqno_t& ls)
+        void finalize(wsrep_seqno_t const ls, int const pa_range)
         {
-            header_.set_last_seen(ls);
+            header_.finalize(ls, pa_range);
         }
 
-        void set_preordered (ssize_t pa_range)
+        /* Serializes wiriteset into a single buffer (for unit test purposes)
+         * set last_seen to -1 if ws was explicitly finalized */
+        void serialize(std::vector<gu::byte_t>& ret,
+                       const wsrep_uuid_t&      source,
+                       const wsrep_conn_id_t&   conn,
+                       const wsrep_trx_id_t&    trx,
+                       const wsrep_seqno_t      last_seen,
+                       const int                pa_range = -1)
+        {
+            WriteSetNG::GatherVector out;
+            size_t const out_size(gather(source, conn, trx, out));
+            finalize(last_seen, pa_range);
+
+            ret.clear(); ret.reserve(out_size);
+
+            /* concatenate all out buffers into ret */
+            for (size_t i(0); i < out->size(); ++i)
+            {
+                const gu::byte_t* ptr
+                    (static_cast<const gu::byte_t*>(out[i].ptr));
+                ret.insert (ret.end(), ptr, ptr + out[i].size);
+            }
+        }
+
+        void finalize_preordered (ssize_t pa_range)
         {
             assert (pa_range >= 0);
 
@@ -598,11 +637,7 @@ namespace galera
              * 0 meaning failed certification. */
             pa_range++;
 
-            /* cap PA range by maximum we can represent */
-            if (gu_unlikely(pa_range > WriteSetNG::MAX_PA_RANGE))
-                pa_range = WriteSetNG::MAX_PA_RANGE;
-
-            header_.set_preordered(pa_range + 1);
+            header_.finalize_preordered(pa_range);
         }
 
     private:
@@ -697,23 +732,34 @@ namespace galera
               check_ (false)
         {}
 
-        /* WriteSetIn(buf) == WriteSetIn() + read_buf(buf) */
-        void read_buf (const gu::Buf& buf, ssize_t const st = SIZE_THRESHOLD)
+        void read_header (const gu::Buf& buf)
         {
             assert (0 == size_);
             assert (false == check_);
 
             header_.read_buf (buf);
             size_ = buf.size;
+        }
+
+        /*
+         * WriteSetIn(buf) == WriteSetIn() + read_buf(buf)
+         *
+         * @param st threshold at which launch dedicated thread for checksumming
+         *           0 - no checksumming
+         */
+        void read_buf (const gu::Buf& buf, ssize_t const st = SIZE_THRESHOLD)
+        {
+            read_header (buf);
             gu_trace(init(st));
         }
 
-        void read_buf (const gu::byte_t* const ptr, ssize_t const len)
+        void read_buf (const void* const ptr, ssize_t const len,
+                       ssize_t const st = SIZE_THRESHOLD)
         {
             assert (ptr != NULL);
             assert (len >= 0);
-            gu::Buf tmp = { ptr, len };
-            read_buf (tmp);
+            gu::Buf tmp = { static_cast<const gu::byte_t*>(ptr), len };
+            read_buf (tmp, st);
         }
 
         ~WriteSetIn ()
@@ -727,14 +773,21 @@ namespace galera
             delete annt_;
         }
 
-        size_t        size()      const { return size_;               }
+        WriteSetNG::Version version()   const { return header_.version(); }
+
+        ssize_t       size()      const { return size_;               }
         uint16_t      flags()     const { return header_.flags();     }
-        bool          is_toi()    const
-        { return flags() & WriteSetNG::F_TOI; }
+        bool          is_toi()    const { return flags() & WriteSetNG::F_TOI; }
         bool          pa_unsafe() const
         { return flags() & WriteSetNG::F_PA_UNSAFE; }
         int           pa_range()  const { return header_.pa_range();  }
-        bool          certified() const { return header_.pa_range();  }
+        bool          certified() const
+        {
+            if (gu_likely(version() >= WriteSetNG::VER5))
+                return (flags() & WriteSetNG::F_CERTIFIED);
+            else
+                return (pa_range()); // VER3
+        }
         wsrep_seqno_t last_seen() const { return header_.last_seen(); }
         wsrep_seqno_t seqno()     const { return header_.seqno();     }
         long long     timestamp() const { return header_.timestamp(); }
@@ -770,7 +823,7 @@ namespace galera
             return (data_.get_checksum());
         }
 
-        void set_seqno(const wsrep_seqno_t& seqno, ssize_t pa_range)
+        void set_seqno(wsrep_seqno_t const seqno, int pa_range)
         {
             assert (seqno    >  0);
             assert (pa_range >= 0);
