@@ -20,7 +20,7 @@
 #endif
 
 #include "wsrep_params.hpp"
-#include "event_service.hpp"
+#include "gu_event_service.hpp"
 #include "wsrep_config_service.h"
 
 #include <cassert>
@@ -1705,12 +1705,12 @@ extern "C" void wsrep_deinit_allowlist_service_v1()
 extern "C"
 int wsrep_init_event_service_v1(wsrep_event_service_v1_t *event_service)
 {
-    return galera::EventService::init_v1(event_service);
+    return gu::EventService::init_v1(event_service);
 }
 
 extern "C" void wsrep_deinit_event_service_v1()
 {
-    galera::EventService::deinit_v1();
+    gu::EventService::deinit_v1();
 }
 
 static int map_parameter_flags(int flags)
@@ -1726,6 +1726,8 @@ static int map_parameter_flags(int flags)
       ret |= WSREP_PARAM_TYPE_INTEGER;
     if (flags & gu::Config::Flag::type_double)
       ret |= WSREP_PARAM_TYPE_DOUBLE;
+    if (flags & gu::Config::Flag::type_duration)
+      ret |= WSREP_PARAM_TYPE_DOUBLE;
     return ret;
 }
 
@@ -1742,13 +1744,30 @@ static int wsrep_parameter_init(wsrep_parameter& wsrep_param,
         ret = gu_str2bool(param.value().c_str(), &wsrep_param.value.as_bool);
         break;
     case gu::Config::Flag::type_integer:
-        long long llval;
-        ret = gu_str2ll(param.value().c_str(), &llval);
-        wsrep_param.value.as_integer = llval;
+    {
+        long long tmp;
+        ret = gu_str2ll(param.value().c_str(), &tmp);
+        wsrep_param.value.as_integer = tmp;
         break;
+    }
     case gu::Config::Flag::type_double:
         ret = gu_str2dbl(param.value().c_str(), &wsrep_param.value.as_double);
         break;
+    case gu::Config::Flag::type_duration:
+    {
+        try
+        {
+            // durations are mapped to doubles
+            wsrep_param.value.as_double
+                = to_double(gu::datetime::Period(param.value()));
+        }
+        catch (...)
+        {
+            assert(0);
+            return 1;
+        }
+        break;
+    }
     default:
         assert((param.flags() & gu::Config::Flag::type_mask) == 0);
         wsrep_param.value.as_string = param.value().c_str();
@@ -1796,12 +1815,16 @@ extern "C"
 int wsrep_init_config_service_v1(wsrep_config_service_v1_t *config_service)
 {
     config_service->get_parameters = get_parameters;
+    // Deprecation checks will be done by application which uses
+    // the service.
+    gu::Config::disable_deprecation_check();
     return WSREP_OK;
 }
 
 extern "C"
 void wsrep_deinit_config_service_v1()
 {
+    gu::Config::enable_deprecation_check();
 }
 
 extern "C"
