@@ -92,8 +92,8 @@ public:
                                                  Defaults::PcWeight),
                                       0, 0xff)),
         rst_view_         (),
-        sync_param_mutex_ (),
-        sync_param_cond_  (),
+        sync_param_mutex_ (0),
+        sync_param_cond_  (0),
         param_sync_set_   (0)
 
     {
@@ -146,7 +146,7 @@ public:
     { NodeMap::value(self_i_).set_weight(weight); }
 
 
-    class SMMap : public Map<const UUID, Message> { };
+    class SMMap : public Map<UUID, Message> { };
 
     const View& current_view() const { return current_view_; }
 
@@ -224,8 +224,8 @@ private:
         mtu_              (std::numeric_limits<int32_t>::max()),
         weight_           (),
         rst_view_         (),
-        sync_param_mutex_ (),
-        sync_param_cond_  (),
+        sync_param_mutex_ (0),
+        sync_param_cond_  (0),
         param_sync_set_   (0)
     { }
 
