@@ -1,11 +1,11 @@
 #!/usr/bin/env bash
 
-set -eu
+set -eux
 
 # $Id$
 
 # Galera library version
-VERSION="26.4.13"
+VERSION="26.4.16"
 
 get_cores()
 {
@@ -52,6 +52,10 @@ Options:
 EOF
 }
 
+# Scons executable on RHEL-8 is scons-3
+SCONSBIN=scons
+[[ -n "$(which scons-3)" ]] && SCONSBIN=scons-3
+
 OS=$(uname)
 # disable building vsbes by default
 DISABLE_VSBES=${DISABLE_VSBES:-"yes"}
@@ -99,7 +103,7 @@ elif [ "$OS" == "FreeBSD" ]; then
   EXTRA_SYSROOT=/usr/local
 fi
 
-which dpkg >/dev/null 2>&1 && DEBIAN=${DEBIAN:-1} || DEBIAN=${DEBIAN:-0}
+which dpkg-buildpackage >/dev/null 2>&1 && DEBIAN=${DEBIAN:-1} || DEBIAN=${DEBIAN:-0}
 
 if [ "$OS" == "FreeBSD" ]; then
     CC=${CC:-"clang"}
@@ -109,12 +113,6 @@ else
     CXX=${CXX:-"g++"}
 fi
 
-if ccache -V > /dev/null 2>&1
-then
-    echo "$CC"  | grep "ccache" > /dev/null || CC="ccache $CC"
-    echo "$CXX" | grep "ccache" > /dev/null || CXX="ccache $CXX"
-fi
-
 # MariaDB Centos5 & SLES-SP1 buildbot images require LD_LIBRARY_PATH to
 # /usr/local/lib
 if [ -r /etc/redhat-release ]; then
@@ -319,11 +317,11 @@ build_packages()
 
     set +e
     if [ $DEBIAN -ne 0 ]; then # build DEB
-        debian_version="$(lsb_release -sc)"
-
-        # Adjust compat for older platforms
-        test "$debian_version" != "lucid" || echo 7 > debian/compat
-        test "$debian_version" != "squeeze" || echo 8 > debian/compat
+        source /etc/os-release
+        # sid doesn't have VERSION_ID set
+        version=${VERSION_ID:-sid}
+        # remove . from version e.g. 22.04
+        debian_version=${ID:0:3}${version/.}
 
         dch -m -D "$debian_version" --force-distribution -v "$GALERA_VER-$debian_version" "Version upgrade"
         # -d : Do not check build dependencies and conflicts.
@@ -340,7 +338,6 @@ build_packages()
         fi
         RET=$?
     fi
-    local RET=$?
 
     set -e
 
@@ -409,9 +406,7 @@ then
 fi
 if [ "$PACKAGE" == "yes" -a $DEBIAN -ne 0 ]
 then
-    # Skip extra build step when building package for debian
-    echo "Reset env for Debian package build"
-    unset CC CXX LD_LIBRARY_PATH CPPFLAGS CFLAGS CXXFLAGS
+    echo "Debian package build"
 elif [ "$CMAKE" == "yes" ] # Build using CMake
 then
     cmake_args="$CMAKE_OPTS -DGALERA_REVISION=$GALERA_REV"
@@ -452,12 +447,12 @@ then
 
     if [ "$SCRATCH" == "yes" ]
     then
-        scons -Q -c --conf=force $scons_args $SCONS_OPTS
+        ${SCONSBIN} -Q -c --conf=force $scons_args $SCONS_OPTS
     fi
 
     if [ "$SKIP_BUILD" != "yes" ]
     then
-        scons $scons_args -j $JOBS $SCONS_OPTS
+        ${SCONSBIN} $scons_args -j $JOBS $SCONS_OPTS
     fi
 
 elif test "$SKIP_BUILD" == "no"; then # Build using autotools
